from datetime import datetime
from dataclasses import dataclass
from typing import List, Optional
from time import sleep

from requests import codes

from .constants import _LOGGER, MAP_URL
from .types import ChargingSessionUpdate, PowerUtility
from .exceptions import ChargePointCommunicationException


def _modify(
    client: "ChargePoint",  # noqa: F821
    action: str,
    device_id: int,
    port_number: int = 1,
    session_id: int = 0,
    max_retry: int = 30,
) -> Optional[int]:
    if action not in ["start", "stop"]:
        raise AttributeError(f"Invalid action: {action}")

    request = {
        "deviceData": client.device_data,
        "deviceId": device_id,
    }

    if action == "stop":
        request["portNumber"] = port_number
        request["sessionId"] = session_id

    action_path = {
        "start": "startsession",
        "stop": "stopSession",
    }

    response = client.session.post(
        f"https://account.chargepoint.com/account/v1/driver/station/{action_path[action]}",
        json=request,
    )

    if response.status_code != codes.ok:
        _LOGGER.error(
            "Failed to send command to station! status_code=%s err=%s",
            response.status_code,
            response.text,
        )
        raise ChargePointCommunicationException(
            response=response, message=f"Failed to {action} ChargePoint session."
        )

    action_status = response.json()
    ack_id = action_status["ackId"]

    for i in range(max_retry):  # pragma: no cover
        _LOGGER.debug(
            "Checking station modification status. (Attempt %d/%d)",
            i + 1,
            max_retry,
        )
        request = {
            "deviceData": client.device_data,
            "ackId": ack_id,
            "action": f"{action}_session",
        }
        response = client.session.post(
            "https://account.chargepoint.com/account/v1/driver/station/session/ack",
            json=request,
        )
        if response.status_code == codes.ok:
            _LOGGER.info("Successfully confirmed %s command.", action)

            # If we just started a new session, return the new ID.
            if action == "start":
                return response.json().get("sessionId")
            # Otherwise, just return.
            return

        if i == max_retry - 1:
            _LOGGER.error(
                "Failed to confirm station modification! status_code=%s err=%s",
                response.status_code,
                response.text,
            )
            raise ChargePointCommunicationException(
                response=response,
                message=f"Session failed to {action} in time allotted.",
            )
        sleep(1)


@dataclass
class ChargingSession:
    session_id: int
    start_time: datetime

    # Device Information
    device_id: int
    device_name: str
    charging_state: str
    charging_time: int
    energy_kwh: float
    miles_added: float
    miles_added_per_hour: float
    outlet_number: int
    port_level: int
    power_kw: float
    purpose: str

    # Payment
    currency_iso_code: str
    payment_completed: bool
    payment_type: str
    pricing_spec_id: int
    total_amount: float

    # API / Misc.
    api_flag: bool
    enable_stop_charging: bool
    has_charging_receipt: bool
    has_utility_info: bool
    is_home_charger: bool
    is_purpose_finalized: bool
    last_update_data_timestamp: datetime
    stop_charge_supported: bool

    # Owner / Location
    company_id: int
    company_name: str
    latitude: float
    longitude: float
    address: str
    city: str
    state_name: str
    country: str
    zipcode: str

    # Session Update History
    update_data: List[ChargingSessionUpdate]
    update_period: int

    utility: Optional[PowerUtility]

    def __init__(
        self, session_id: int, client: "ChargePoint", *args, **kwargs  # noqa: F821
    ):
        super().__init__(*args, **kwargs)
        self._client = client
        self.session_id = session_id

        self._get()

    def _get(self) -> None:
        _LOGGER.debug("Getting session information for session %s", self.session_id)
        response = None

        # There is some level of eventual consistency where with sessions
        # that are started and then immediately retrieved, this normally
        # becomes consistent within a few seconds, so we will retry this
        # call up to 10 times.
        for attempt in range(1, 11):  # pragma: no cover
            # Today on "Every Internal API is Weird as Hell"...
            # I present to you: passing a JSON blob as a URL parameter.
            response = self._client.session.get(
                f'{MAP_URL}?{{"user_id":{self._client.user_id},"charging_status":{{"mfhs":{{}},"session_id":{self.session_id}}}}}'
            )

            if response.status_code != codes.ok:
                raise ChargePointCommunicationException(
                    response=response, message="Failed to get charging session data."
                )

            json = response.json()
            status = json.get("charging_status", {})
            # Failed calls here have been NullPointers from their API
            # directly returned to the user.
            # There have also been a couple random "this ID doesn't exist"
            # logical errors as well, so lets just cover all the bases.
            error = (
                # Sometimes this key isn't even returned.
                "charging_status" not in json.keys()
                or
                # Logical errors from withing the API returned here.
                "error_message" in status.keys()
                or
                # Raw Java errors are returned here.
                "error" in status.keys()
            )
            if error and attempt < 10:
                _LOGGER.warning("Failed to retrieve session. Attempt (%d/10)", attempt)
                _LOGGER.debug("%s", json)
                sleep(1)
                continue
            elif error:
                raise ChargePointCommunicationException(
                    response=response, message="Failed to get charging session data."
                )
            else:
                break

        _LOGGER.debug("Passed retry loop: %s", response.json())
        status = response.json()["charging_status"]

        self.start_time = datetime.fromtimestamp(status["start_time"] / 1000)
        self.device_id = status["device_id"]
        self.device_name = status["device_name"]
        self.charging_state = status["current_charging"]
        self.charging_time = status["charging_time"]
        self.energy_kwh = status["energy_kwh"]
        self.miles_added = status["miles_added"]
        self.miles_added_per_hour = status.get("miles_added_per_hour", 0.0)
        self.outlet_number = status["outlet_number"]
        self.port_level = status["port_level"]
        self.power_kw = status["power_kw"]
        self.purpose = status["purpose"]
        self.currency_iso_code = status["currency_iso_code"]
        self.payment_completed = status["payment_completed"]
        self.payment_type = status["payment_type"]
        self.pricing_spec_id = status["pricing_spec_id"]
        self.total_amount = status["total_amount"]
        self.api_flag = status["api_flag"]
        self.enable_stop_charging = status["enable_stop_charging"]
        self.has_charging_receipt = status["has_charging_receipt"]
        self.has_utility_info = status["has_utility_info"]
        self.is_home_charger = status["is_home_charger"]
        self.is_purpose_finalized = status["is_purpose_finalized"]
        self.last_update_data_timestamp = datetime.fromtimestamp(
            status["last_update_data_timestamp"] / 1000
        )
        self.stop_charge_supported = status["stop_charge_supported"]
        self.company_id = status["company_id"]
        self.company_name = status["company_name"]
        self.latitude = status["lat"]
        self.longitude = status["lon"]
        self.address = status["address1"]
        self.city = status["city"]
        self.state_name = status["state_name"]
        self.country = status["country"]
        self.zipcode = status["zipcode"]
        self.update_data = [
            ChargingSessionUpdate.from_json(update) for update in status["update_data"]
        ]
        self.update_period = status["update_period"]

        self.utility = None
        utility = status.get("utility")
        if utility:
            self.utility = PowerUtility.from_json(utility)

    def stop(self, max_retry: int = 30) -> None:
        _modify(
            client=self._client,
            action="stop",
            device_id=self.device_id,
            port_number=self.outlet_number,
            session_id=self.session_id,
            max_retry=max_retry,
        )

    @classmethod
    def start(
        cls, device_id: int, client: "ChargePoint", max_retry: int = 30  # noqa: F821
    ):
        session_id = _modify(
            client=client, action="start", device_id=device_id, max_retry=max_retry
        )
        # So, after wayyy too much trial and error, I noticed that the "sessionId"
        # returned by the start session API is significantly higher than normal
        # session IDs... I have no clue what it means, so we are just going to
        # get the correct session ID from the status API.
        status = client.get_user_charging_status()
        if session_id and status:  # pragma: no cover
            return cls(session_id=status.session_id, client=client)



from requests import Response


class ChargePointBaseException(Exception):
    """
    Root exception for all module raised errors.
    """


class ChargePointCommunicationException(ChargePointBaseException):
    """
    Parent class for exceptions that involve communication
    with the ChargePoint API.
    """

    def __init__(self, response: Response, message: str):
        self.response = response
        self.message = message
        super().__init__(self.message)


class ChargePointLoginError(ChargePointCommunicationException):
    """
    Login failed.
    """


class ChargePointInvalidSession(ChargePointCommunicationException):
    """
    Login expired.
    """



from dataclasses import dataclass
from datetime import datetime
from typing import List

from .constants import _LOGGER


@dataclass
class ElectricVehicle:
    make: str
    model: str
    primary_vehicle: bool
    color: str
    image_url: str
    year: int
    charging_speed: int
    dc_charging_speed: int

    @classmethod
    def from_json(cls, json: dict):
        return cls(
            make=json["make"]["name"],
            model=json["model"]["name"],
            primary_vehicle=json["primaryVehicle"],
            color=json["modelYearColor"]["colorName"],
            image_url=json["modelYearColor"]["imageUrl"],
            year=json["modelYear"]["year"],
            charging_speed=json["modelYear"]["chargingSpeed"],
            dc_charging_speed=json["modelYear"]["dcChargingSpeed"],
        )


@dataclass
class ChargePointUser:
    email: str
    evatar_url: str
    family_name: str
    full_name: str
    given_name: str
    phone: str
    phone_country_id: int
    user_id: int
    username: str

    @classmethod
    def from_json(cls, json: dict):
        return cls(
            email=json["email"],
            evatar_url=json["evatarUrl"],
            family_name=json["familyName"],
            full_name=json["fullName"],
            given_name=json["givenName"],
            phone=json["phone"],
            phone_country_id=json["phoneCountryId"],
            user_id=json["userId"],
            username=json["username"],
        )


@dataclass
class AccountBalance:
    account_number: str
    account_state: str
    amount: str
    currency: str

    @classmethod
    def from_json(cls, json: dict):
        balance = json["balance"]
        return cls(
            account_number=json["accountNumber"],
            account_state=json["accountState"],
            amount=balance["amount"],
            currency=balance["currency"],
        )


@dataclass
class ChargePointAccount:
    user: ChargePointUser
    account_balance: AccountBalance

    @classmethod
    def from_json(cls, json: dict):
        user = ChargePointUser.from_json(json["user"])
        balance = AccountBalance.from_json(json["accountBalance"])
        return cls(user=user, account_balance=balance)


@dataclass
class HomeChargerStatus:
    charger_id: int
    brand: str
    plugged_in: bool
    connected: bool
    charging_status: str  # "AVAILABLE", "CHARGING", "NOT_CHARGING"
    last_connected_at: datetime
    reminder_enabled: bool
    reminder_time: str
    model: str
    mac_address: str

    @classmethod
    def from_json(cls, charger_id: int, json: dict):
        return cls(
            charger_id=charger_id,
            brand=json["brand"],
            plugged_in=json["is_plugged_in"],
            connected=json["is_connected"],
            charging_status=json["charging_status"],
            last_connected_at=datetime.fromtimestamp(json["last_connected_at"] / 1000),
            reminder_enabled=json["is_reminder_enabled"],
            reminder_time=json["plug_in_reminder_time"],
            model=json["model"],
            mac_address=json["mac_address"],
        )


@dataclass
class HomeChargerTechnicalInfo:
    model: str
    serial_number: str
    mac_address: str
    software_version: str
    last_ota_update: datetime
    device_ip: str
    last_connected_at: datetime
    is_stop_charge_supported: bool

    @classmethod
    def from_json(cls, json: dict):
        return cls(
            model=json["model_number"],
            serial_number=json["serial_number"],
            mac_address=json["wifi_mac"],
            software_version=json["software_version"],
            last_ota_update=datetime.fromtimestamp(json["last_ota_update"] / 1000),
            device_ip=json["device_ip"],
            last_connected_at=datetime.fromtimestamp(json["last_connected_at"] / 1000),
            is_stop_charge_supported=json["is_stop_charge_supported"],
        )


@dataclass
class ChargePointStation:
    id: int
    name: str
    latitude: float
    longitude: float

    @classmethod
    def from_json(cls, json):
        return cls(
            id=json["deviceId"],
            name=json["name"],
            latitude=json["lat"],
            longitude=json["lon"],
        )


@dataclass
class UserChargingStatus:
    session_id: int
    start_time: datetime
    state: str  # "in_use", "fully_charged"
    stations: List[ChargePointStation]

    @classmethod
    def from_json(cls, json: dict):
        status = json["charging"]
        state = status.get("state", "unknown")
        if state == "unknown":
            _LOGGER.warning(
                "Charging status returned without a state. "
                + "This is normally due to the eventually consistent "
                + "nature of the session API."
            )
        return cls(
            session_id=status["sessionId"],
            start_time=datetime.fromtimestamp(status["startTimeUTC"]),
            state=state,
            stations=[
                ChargePointStation.from_json(station) for station in status["stations"]
            ],
        )


@dataclass
class ChargingSessionUpdate:
    energy_kwh: float
    power_kw: float
    timestamp: datetime

    @classmethod
    def from_json(cls, json: dict):
        return cls(
            energy_kwh=json["energy_kwh"],
            power_kw=json["power_kw"],
            timestamp=datetime.fromtimestamp(json["timestamp"] / 1000),
        )


@dataclass
class PowerUtilityPlan:
    code: str
    id: int
    is_ev_plan: bool
    name: str

    @classmethod
    def from_json(cls, json: dict):
        return cls(
            code=json["code"],
            id=json["id"],
            is_ev_plan=json["is_ev_plan"],
            name=json["name"],
        )


@dataclass
class PowerUtility:
    id: int
    name: str
    plans: List[PowerUtilityPlan]

    @classmethod
    def from_json(cls, json: dict):
        return cls(
            id=json["id"],
            name=json["name"],
            plans=[PowerUtilityPlan.from_json(plan) for plan in json["plans"]],
        )



from uuid import uuid4
from typing import List, Optional
from functools import wraps

from requests import Session, codes, post

from .types import (
    ChargePointAccount,
    ElectricVehicle,
    HomeChargerStatus,
    HomeChargerTechnicalInfo,
    UserChargingStatus,
)
from .exceptions import (
    ChargePointLoginError,
    ChargePointCommunicationException,
    ChargePointBaseException,
    ChargePointInvalidSession,
)
from .session import ChargingSession
from .constants import _LOGGER, V5_URL, MAP_URL


def _dict_for_query(device_data: dict) -> dict:
    """
    GET requests send device data as a nested object.
    To avoid storing the device data block in two
    formats, we are just going to compute the flat
    dictionary.
    """
    return {f"deviceData[{key}]": value for key, value in device_data.items()}


def _require_login(func):
    @wraps(func)
    def check_login(*args, **kwargs):
        self = args[0]
        if not self._logged_in:
            raise RuntimeError("Must login to use ChargePoint API")
        try:
            return func(*args, **kwargs)
        except ChargePointCommunicationException as exc:
            if exc.response.status_code == codes.unauthorized:
                raise ChargePointInvalidSession(
                    exc.response, "Session token has expired. Please login again!"
                ) from exc
            else:
                raise

    return check_login


class ChargePoint:
    def __init__(
        self,
        username: str,
        password: str,
        session_token: str = "",
        app_version: str = "5.91.0",
    ):
        self._session = Session()
        self._app_version = app_version
        self._device_data = {
            "appId": "com.coulomb.ChargePoint",
            "manufacturer": "Apple",
            "model": "iPhone",
            "notificationId": "",
            "notificationIdType": "",
            "type": "IOS",
            "udid": str(uuid4()),
            "version": app_version,
        }
        self._device_query_params = _dict_for_query(self._device_data)
        self._user_id = None
        self._logged_in = False
        self._session_token = None

        if session_token:
            self._set_session_token(session_token)
            self._logged_in = True
            try:
                account: ChargePointAccount = self.get_account()
                self._user_id = str(account.user.user_id)
                return
            except ChargePointCommunicationException:
                _LOGGER.warning(
                    "Provided session token is expired, will attempt to re-login"
                )
                self._logged_in = False

        self.login(username, password)

    @property
    def user_id(self) -> Optional[str]:
        return self._user_id

    @property
    def session(self) -> Session:
        return self._session

    @property
    def session_token(self) -> Optional[str]:
        return self._session_token

    @property
    def device_data(self) -> dict:
        return self._device_data

    def login(self, username: str, password: str) -> None:
        """
        Create a session and login to ChargePoint
        :param username: Account username
        :param password: Account password
        """
        login_url = (
            "https://account.chargepoint.com/account/v2/driver/profile/account/login"
        )
        headers = {
            "User-Agent": f"com.coulomb.ChargePoint/{self._app_version} CFNetwork/1329 Darwin/21.3.0"
        }
        request = {
            "deviceData": self._device_data,
            "username": username,
            "password": password,
        }
        _LOGGER.debug("Attempting client login with user: %s", username)
        login = post(login_url, json=request, headers=headers)
        _LOGGER.debug(login.cookies.get_dict())
        _LOGGER.debug(login.headers)

        if login.status_code == codes.ok:
            req = login.json()
            self._user_id = req["user"]["userId"]
            _LOGGER.debug("Authentication success! User ID: %s", self._user_id)
            self._set_session_token(req["sessionId"])
            self._logged_in = True
            return

        _LOGGER.error(
            "Failed to get account information! status_code=%s err=%s",
            login.status_code,
            login.text,
        )
        raise ChargePointLoginError(login, "Failed to authenticate to ChargePoint!")

    def logout(self):
        response = self._session.post(
            "https://account.chargepoint.com/account/v1/driver/profile/account/logout",
            json={"deviceData": self._device_data},
        )

        if response.status_code != codes.ok:
            raise ChargePointCommunicationException(
                response=response, message="Failed to log out!"
            )

        self._session.headers = {}
        self._session.cookies.clear_session_cookies()
        self._session_token = None
        self._logged_in = False

    def _set_session_token(self, session_token: str):
        try:
            self._session.headers = {
                "cp-session-type": "CP_SESSION_TOKEN",
                "cp-session-token": session_token,
                # Data:       |------------------Token Data------------------||---?---||-Reg-|
                # Session ID: rAnDomBaSe64EnCodEdDaTaToKeNrAnDomBaSe64EnCodEdD#D???????#RNA-US
                "cp-region": session_token.split("#R")[1],
                "user-agent": "ChargePoint/225 (iPhone; iOS 15.3; Scale/3.00)",
            }
        except IndexError:
            raise ChargePointBaseException("Invalid session token format.")

        self._session_token = session_token
        self._session.cookies.set("coulomb_sess", session_token)

    @_require_login
    def get_account(self) -> ChargePointAccount:
        _LOGGER.debug("Getting ChargePoint Account Details")
        response = self._session.get(
            "https://account.chargepoint.com/account/v1/driver/profile/user",
            params=self._device_query_params,
        )

        if response.status_code != codes.ok:
            _LOGGER.error(
                "Failed to get account information! status_code=%s err=%s",
                response.status_code,
                response.text,
            )
            raise ChargePointCommunicationException(
                response=response, message="Failed to get user information."
            )

        account = response.json()
        return ChargePointAccount.from_json(account)

    @_require_login
    def get_vehicles(self) -> List[ElectricVehicle]:
        _LOGGER.debug("Listing vehicles")
        response = self._session.get(
            "https://account.chargepoint.com/account/v1/driver/vehicle",
            params=self._device_query_params,
        )

        if response.status_code != codes.ok:
            _LOGGER.error(
                "Failed to list vehicles! status_code=%s err=%s",
                response.status_code,
                response.text,
            )
            raise ChargePointCommunicationException(
                response=response, message="Failed to retrieve EVs."
            )

        evs = response.json()
        return [ElectricVehicle.from_json(ev) for ev in evs]

    @_require_login
    def get_home_chargers(self) -> List[int]:
        _LOGGER.debug("Searching for registered pandas")
        get_pandas = {"user_id": self.user_id, "get_pandas": {"mfhs": {}}}
        response = self._session.post(V5_URL, json=get_pandas)

        if response.status_code != codes.ok:
            _LOGGER.error(
                "Failed to get home chargers! status_code=%s err=%s",
                response.status_code,
                response.text,
            )
            raise ChargePointCommunicationException(
                response=response, message="Failed to retrieve Home Flex chargers."
            )

        # {"get_pandas":{"device_ids":[12345678]}}
        pandas = response.json()["get_pandas"]["device_ids"]
        _LOGGER.debug(
            "Discovered %d connected pandas: %s",
            len(pandas),
            ",".join([str(p) for p in pandas]),
        )
        return pandas

    @_require_login
    def get_home_charger_status(self, charger_id: int) -> HomeChargerStatus:
        _LOGGER.debug("Getting status for panda: %s", charger_id)
        get_status = {
            "user_id": self.user_id,
            "get_panda_status": {"device_id": charger_id, "mfhs": {}},
        }
        response = self._session.post(V5_URL, json=get_status)

        if response.status_code != codes.ok:
            _LOGGER.error(
                "Failed to determine home charger status! status_code=%s err=%s",
                response.status_code,
                response.text,
            )
            raise ChargePointCommunicationException(
                response=response, message="Failed to get home charger status."
            )

        status = response.json()

        _LOGGER.debug(status)

        return HomeChargerStatus.from_json(
            charger_id=charger_id, json=status["get_panda_status"]
        )

    @_require_login
    def get_home_charger_technical_info(
        self, charger_id: int
    ) -> HomeChargerTechnicalInfo:
        _LOGGER.debug("Getting tech info for panda: %s", charger_id)
        get_tech_info = {
            "user_id": self.user_id,
            "get_station_technical_info": {"device_id": charger_id, "mfhs": {}},
        }

        response = self._session.post(V5_URL, json=get_tech_info)

        if response.status_code != codes.ok:
            _LOGGER.error(
                "Failed to determine home charger tech info! status_code=%s err=%s",
                response.status_code,
                response.text,
            )
            raise ChargePointCommunicationException(
                response=response, message="Failed to get home charger tech info."
            )

        status = response.json()

        _LOGGER.debug(status)

        return HomeChargerTechnicalInfo.from_json(
            json=status["get_station_technical_info"]
        )

    @_require_login
    def get_user_charging_status(self) -> Optional[UserChargingStatus]:
        _LOGGER.debug("Checking account charging status")
        request = {"deviceData": self._device_data, "user_status": {"mfhs": {}}}
        response = self._session.post(MAP_URL, json=request)

        if response.status_code != codes.ok:
            _LOGGER.error(
                "Failed to get account charging status! status_code=%s err=%s",
                response.status_code,
                response.text,
            )
            raise ChargePointCommunicationException(
                response=response, message="Failed to get user charging status."
            )

        status = response.json()
        if not status["user_status"]:
            _LOGGER.debug("No user status returned, assuming not charging.")
            return None

        _LOGGER.debug("Raw status: %s", status)

        return UserChargingStatus.from_json(status["user_status"])

    @_require_login
    def get_charging_session(self, session_id: int) -> ChargingSession:
        return ChargingSession(session_id=session_id, client=self)

    @_require_login
    def start_charging_session(
        self, device_id: int, max_retry: int = 30
    ) -> ChargingSession:

        return ChargingSession.start(
            device_id=device_id, client=self, max_retry=max_retry
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['python_chargepoint']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.27.1,<3.0.0']

setup_kwargs = {
    'name': 'python-chargepoint',
    'version': '1.2.1',
    'description': 'A simple, Pythonic wrapper for the ChargePoint API.',
    'long_description': None,
    'author': 'Marc Billow',
    'author_email': 'mbillow@users.noreply.github.compoetry',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import sys
import logging
from getpass import getpass

from .client import ChargePoint
from .constants import _LOGGER
from .exceptions import ChargePointLoginError

if __name__ == "__main__":
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(levelname)s: %(message)s")
    stream_handler.setFormatter(formatter)
    _LOGGER.addHandler(stream_handler)
    _LOGGER.setLevel(logging.DEBUG)

    username = input("ChargePoint Username: ")
    password = getpass("Password: ")

    try:
        client = ChargePoint(
            username,
            password,
        )
        print(client.session_token)
    except ChargePointLoginError:
        sys.exit(1)

    print("\n=== Account Information ===")
    acct = client.get_account()
    print(f"Name: {acct.user.full_name}")
    print(f"Balance: {acct.account_balance.amount} {acct.account_balance.currency}")

    print("\n=== Vehicles ===")
    evs = client.get_vehicles()
    for ev in evs:
        print(f"{ev.year} {ev.make} {ev.model}:")
        print(f"  Color: {ev.color}")
        print(f"  Charging Speed: {ev.charging_speed} kW")
        print(f"  DC Charging Speed: {ev.dc_charging_speed} kW")

    home_chargers = client.get_home_chargers()
    chargers_ready_to_charge = []
    if home_chargers:
        print("\n=== Home Charger ===")
    for c in home_chargers:
        panda = client.get_home_charger_status(c)
        tech_info = client.get_home_charger_technical_info(c)

        if panda.charging_status == "AVAILABLE":
            chargers_ready_to_charge.append(panda)

        print(f"{panda.brand} {panda.model}")
        print(f"  Connected: {panda.connected} (Last Seen: {panda.last_connected_at})")
        print(f"  Plugged-In: {panda.plugged_in}")
        print(f"  Status: {panda.charging_status}")
        print(f"  Software Version: {tech_info.software_version}")
        if panda.reminder_enabled:
            print(f"  Reminder: {panda.reminder_time}")

    is_charging = client.get_user_charging_status()
    print("\n=== Account Charging Status ===")
    if is_charging:
        print(f"Charging at {is_charging.stations[0].name}:")
        session = client.get_charging_session(is_charging.session_id)
        print(f"  State: {session.charging_state}")
        print(f"  Miles Added: {session.miles_added}")
        print(f"  Energy Used: {session.energy_kwh} kWh")
        print(f"  Cost: {session.total_amount} {session.currency_iso_code}")

        stop_session = input("End current session? [yes|no]: ")
        if stop_session == "yes":
            session.stop()

        start_session = input("Start a new session? [yes|no]: ")
        if start_session == "yes":
            client.start_charging_session(device_id=session.device_id)
            print(f"Resumed session with new ID: {session.session_id}")

    elif chargers_ready_to_charge:
        for panda in chargers_ready_to_charge:
            start_session = input(
                f"Start a new session on {panda.model} ({panda.charger_id})? [yes|no]: "
            )
            if start_session == "yes":
                client.start_charging_session(panda.charger_id)
